from quingo.utils import ensure_path
from .backend_hub import BackendType
from .if_backend import If_backend
from quingo.core.exe_config import ExeConfig, ExeMode
from symqc.simulator import SymQC


class IfSymQC(If_backend):
    """A functional QCIS simulation backend based on symbolic computation."""

    def __init__(self):
        super().__init__(BackendType.SYMQC)
        self.sim = SymQC()

    def upload_program(self, prog_fn):
        """
        Upload assembly or binary program to the simulator.

        Args:
            prog_fn: the name of the assembly file.
        """
        prog_fn = ensure_path(prog_fn)
        if str(prog_fn).endswith(".qcis"):
            self.sim.compile_file(prog_fn)
        else:
            raise TypeError("The SymQC simulator can only accept QCIS instructions.")

    def execute(self, exe_config: ExeConfig):
        """Execute the given quantum circuit.
        Args:
          - exe_config (ExeConfig): the configuration used to perform simulation by SymQC.

        The number of shots is specified in exe_config.num_shots, which is only valid for
          ExeMode.SimShots.
        """
        if exe_config.mode == ExeMode.SimShots:
            return self.sim.simulate("one_shot", exe_config.num_shots)

        if exe_config.mode == ExeMode.SimFinalResult:
            raw_res = self.sim.simulate("final_state")
            return raw_res

        if exe_config.mode == ExeMode.SimStateVector:
            raw_res = self.sim.simulate("final_state")
            return raw_res["quantum"]

        raise ValueError(
            "Unsupported execution mode ({}) for symqc.".format(exe_config.mode)
        )



from quingo.utils import ensure_path
from .backend_hub import BackendType
from .if_backend import If_backend
from quingo.core.exe_config import ExeConfig, ExeMode
import pyezQ


class XiaoHong(If_backend):
    def __init__(self):
        super().__init__(BackendType.XIAOHONG)
        self.account = None
        self.qcis_circuit = None

    def upload_program(self, prog_fn):
        """
        Upload assembly or binary program to the simulator.

        Args:
            prog_fn: the name of the assembly or binary file.
        """
        prog_fn = ensure_path(prog_fn)
        self.qcis_circuit = prog_fn.open("r").read()

    def execute(self, exe_config: ExeConfig):
        """Execute the given quantum circuit.
        Args:
          - mode (str): the simulation mode to use:
              - "one_shot": the simulation result is a dictionary with each key being a qubit
                  measured, and the value is the outcome of measuring this qubit.
          - num_shots (int): the number of iterations performed in `one_shot` mode.
          - xh_login_key (str): login key to connect XiaoHong
          - xh_machine_name (str): name of machine name to execute qcis
        """
        if not exe_config.mode == ExeMode.RealMachine:
            raise ValueError(
                f"Unsupported execution mode ({exe_config.mode}) for XiaoHong."
            )

        # connect XiaoHong
        self.set_account(exe_config.xh_login_key, exe_config.xh_machine_name)

        # submit job
        print(f"Start execute:")
        print(f"num shots = {exe_config.num_shots}")
        query_id = self.account.submit_job(
            self.qcis_circuit, num_shots=exe_config.num_shots
        )

        # invalid query
        if not query_id:
            raise EnvironmentError("Fail to connect XiaoHong!")

        # read result
        result = self.account.query_experiment(query_id, max_wait_time=360000)
        result = self.format_result(result)
        return result

    def set_account(self, login_key, machine_name):
        self.account = pyezQ.Account(login_key=login_key, machine_name=machine_name)
        print(f"Set account successfully:")
        print(f"   login key = {login_key[0:5]}" + "*" * (len(login_key) - 5))
        print(f"   machine name = {machine_name}")

    def format_result(self, result):
        origin_result = result[0]["results"]
        return {"qubits": origin_result[0], "results": origin_result[1:]}



from pathlib import Path
import subprocess, logging
from quingo.core.compiler_config import get_mlir_path
from quingo.core.quingo_task import Quingo_task
from quingo.core.preparation import gen_main_file
from quingo.core.quingo_logger import quingo_err, get_logger
from quingo.utils import ensure_path
from quingo.backend.qisa import Qisa, get_qisa_name, get_suffix


logger = get_logger((__name__).split(".")[-1])


def compile(task: Quingo_task, params: tuple, qasm_fn: Path = None, config_file=""):
    """Compile the quingo file with given parameters and return the path of
    the generated qasm file.
    """
    logger.setLevel(logging.INFO)
    gen_main_file(task.called_qu_fn, task.called_func, task.cl_entry_fn, params)

    if qasm_fn is None:
        suffix = get_suffix(task.qisa_type)
        qasm_fn = task.cl_entry_fn.with_suffix(suffix)
    else:
        qasm_fn = ensure_path(qasm_fn)

    quingoc_path = Path(get_mlir_path())

    compile_cmd = compose_cl_cmd(task, qasm_fn, quingoc_path, config_file)
    if task.debug_mode:
        logger.info(compile_cmd)
    ret_value = subprocess.run(
        compile_cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
        shell=True,
    )
    if ret_value.stdout != "":
        logger.info(ret_value.stdout.strip())
    if ret_value.stderr != "":
        msg = "Error message from the compiler:\n\t{}".format(ret_value.stderr)
        quingo_err(msg)
        logger.error(msg)

    if ret_value.returncode != 0:  # failure
        raise RuntimeError(
            "failed to compile the Quingo source file: {}".format(task.called_qu_fn)
        )
    else:
        return qasm_fn


def compose_cl_cmd(task: Quingo_task, qasm_fn: Path, quingoc_path: Path, configfile=""):
    qasm_fn = ensure_path(qasm_fn)
    quingoc_path = ensure_path(quingoc_path)

    cl_path = '"{}"'.format(str(quingoc_path.resolve()))
    cl_entry_fn = '"{}"'.format(str(task.cl_entry_fn))

    opt_inc_dirs = " ".join(['-I "{}"'.format(str(dir)) for dir in task.include_dir])
    opt_isa = "--isa={}".format(get_qisa_name(task.qisa_type))

    qubits_info = ""
    if task.qisa_type == Qisa.Quantify:
        qubits_info = '--qubits="{}"'.format(str(task.qubits_info.resolve()))
    opt_qubit_map = qubits_info

    opt_out_fn = '-o "{}"'.format(str(qasm_fn))

    config_fn = '--config-fn="{}"'.format(str(configfile))

    cmd_eles = [
        cl_path,
        cl_entry_fn,
        opt_inc_dirs,
        config_fn,
        opt_isa,
        opt_qubit_map,
        opt_out_fn,
    ]

    compile_cmd = " ".join([ele for ele in cmd_eles if ele.strip() != ""])

    return compile_cmd



from typing import List, Tuple


def reorder_bits(qubit_list, qubit_msmt_res, Ascending=True):
    """
    Given the measurement result `qubit_msmt_res` of `qubit_list`, reorder the bits in the
    result according to the qubit name order.

    Args:
        qubit_list (list): A list of qubit names in the desired order.
        qubit_msmt_res (list): A list of measurement results corresponding to the qubits.
        Ascending (bool, optional): Specifies whether the bits should be reordered in ascending order.
            Defaults to True.

    Returns:
        list: A list of reordered measurement results.

    Example:
        Given qubit list ["Q3", "Q1", "Q4", "Q2"] and `qubit_msmt_res` [1, 1, 0, 0],
        the ordered result is [1, 0, 1, 0], which corresponds to ["Q1", "Q2", "Q3", "Q4"].
    """
    num_qubits = len(qubit_list)

    qubit_result_pair = [
        (qubit_list[i].lower(), qubit_msmt_res[i]) for i in range(num_qubits)
    ]

    qubit_result_pair.sort(key=lambda x: int(x[0][1:]), reverse=not Ascending)

    print(qubit_result_pair)

    return [x[1] for x in qubit_result_pair]


def get_first_non_zero_res(
    shots_result: Tuple[List[str], List[List[int]]],
    qubits: List[str] = None,
    integer_format=True,
    little_endian=True,
):
    """
    This function is used to get the first non-zero result from the measurement results.

    Args:
        shots_result (tuple): A tuple containing two elements: a list of qubit names and a list
          of measurement results.
            An example shots result of a simulation with num_shots = 3:
            (['Q3', 'Q4', 'Q5', 'Q6', 'Q7'],
               [[1, 1, 0, 0, 0],
                [1, 1, 0, 0, 0],
                [1, 1, 0, 0, 0]])

        qubits (list, optional): A list of qubit names. If provided, the function will only consider
          the measurement results of the specified qubits.
            Defaults to None.

        integer_format (bool, optional): Specifies whether the result should be returned as an
          integer or as a list of bits.
            Defaults to True.

        little_endian (bool, optional): Specifies whether the result should be in little-endian
        format. If False, the result is in big-endian format.
            Defaults to True.
            Little-endian: [Q0, Q1, Q2] [1, 0, 0] -> 1
            Big-endian:    [Q0, Q1, Q2] [1, 0, 0] -> 4

    Returns:
        int or list: The first non-zero result in the measurement results. If `integer_format` is
          True, the result is returned as an integer.
        Otherwise, the result is returned as a list of bits.

    Raises:
        ValueError: If `shots_result` is None or if the measurement results are None.
    """
    if shots_result is None:
        raise ValueError("No measure results found")
    names, msmt_res = shots_result

    if msmt_res is None:
        raise ValueError("No measure results found")

    res = msmt_res[0]
    if qubits is not None:
        indices = [names.index(q) for q in qubits]
        res = [res[idx] for idx in indices]

    if little_endian:
        res.reverse()

    if integer_format:
        return int("".join(map(str, res)), 2)
    else:
        return msmt_res[0]



from __future__ import annotations
from typing import Union
from quingo.utils import ensure_path
from pathlib import Path
from quingo.backend.backend_hub import BackendType
from quingo.backend.if_backend import If_backend
from quingo.core.exe_config import ExeConfig, ExeMode
from qualesim.plugin import Loglevel
from qualesim.host import Simulator


class QuaLeSim(If_backend):

    def __init__(self, backend_type=BackendType.QUALESIM_QUANTUMSIM):
        super().__init__(backend_type)
        self.sim = Simulator(stderr_verbosity=Loglevel.OFF)
        if backend_type == BackendType.QUALESIM_QUANTUMSIM:
            self.sim.with_backend("quantumsim", verbosity=Loglevel.OFF)
        elif backend_type == BackendType.QUALESIM_TEQUILA:
            self.sim.with_backend("tequila", verbosity=Loglevel.OFF)
        else:
            raise ValueError(
                "The QuaLeSim backend only supports QUALESIM_QUANTUMSIM and QUALESIM_TEQUILA."
            )

        self.res = None

    def upload_program(self, prog_fn: Union[Path | str]):
        prog_fn = ensure_path(prog_fn)

        if prog_fn.suffix in [".qcis", ".qi"]:
            self.sim.with_frontend(str(prog_fn), verbosity=Loglevel.OFF)
        else:
            raise TypeError(
                "found unsupported file suffix ({}). Currently supported are "
                "'.qcis' (for QCIS) and '.qi' (for QUIET-s)".format(prog_fn.suffix)
            )

    def execute(self, exe_config: ExeConfig):
        if exe_config.mode == ExeMode.SimShots:
            measure_mod = "one_shot"
            try:
                self.sim.simulate()
                res = self.sim.run(
                    measure_mod=measure_mod, num_shots=exe_config.num_shots
                )
                self.sim.stop()

                final_state = res["res"]
                final_state["quantum"] = tuple(final_state["quantum"])
                result = final_state["quantum"]

                return result

            except Exception as e:
                raise ValueError(
                    "error in QuaLeSim simulation with mode ({}): {}".format(
                        exe_config.mode, e
                    )
                )

        if exe_config.mode == ExeMode.SimFinalResult:
            measure_mod = "one_shot"
            try:
                self.sim.simulate()
                res = self.sim.run(
                    measure_mod=measure_mod, num_shots=exe_config.num_shots
                )
                self.sim.stop()
                final_state = res["res"]
                final_state["quantum"] = tuple(final_state["quantum"])
                return final_state["quantum"]

            except Exception as e:
                raise ValueError(
                    "error in QuaLeSim simulation with mode ({}): {}".format(
                        exe_config.mode, e
                    )
                )

        if exe_config.mode == ExeMode.SimStateVector:
            measure_mod = "state_vector"
            try:
                self.sim.simulate()
                res = self.sim.run(measure_mod=measure_mod)
                self.sim.stop()
                final_state = eval(res["res"])
                return final_state["quantum"]

            except Exception as e:
                raise ValueError(
                    "error in QuaLeSim simulation with mode ({}): {}".format(
                        exe_config.mode, e
                    )
                )

        raise ValueError(
            "Unsupported execution mode ({}) for QuaLeSim_QuantumSim.".format(
                exe_config.mode
            )
        )



import enum


class ExeMode(enum.Enum):
    """
    Enumeration representing different result format.

    Attributes:
        SimFinalResult: for obtaining the final result.
        SimStateVector: Simulation mode for obtaining the state vector.
        SimMatrix: Simulation mode for obtaining the matrix representation.
        RealMachine: Mode for executing on a real quantum machine.
    """

    SimShots = enum.auto()
    SimFinalResult = enum.auto()
    SimStateVector = enum.auto()
    SymbolicStateVector = enum.auto()
    RealMachine = enum.auto()


class ExeConfig:
    def __init__(
        self,
        mode: ExeMode = ExeMode.SimShots,
        num_shots: int = 1,
        xh_login_key: str = None,  # use for connecting XIAOHONG
        xh_machine_name: str = None,  # use for connecting XIAOHONG
    ):
        self.mode = mode
        self.num_shots = num_shots
        self.xh_login_key = xh_login_key
        self.xh_machine_name = xh_machine_name

    def __str__(self) -> str:
        return str(self.mode)



"""The task of this module is to prepare the Quingo source file for the compilation.
"""

import re
from pathlib import Path
import logging
from quingo.core.quingo_logger import get_logger
import quingo.core.data_transfer as dt

logger = get_logger((__name__).split(".")[-1])


def remove_comment(qu_src):
    search_annotation_string = r"\/\/.*\n"
    return re.sub(search_annotation_string, "", qu_src)


def get_ret_type(qu_src: str, qg_func_name: str):
    """This function read the Quingo source file to retrieve the return
    type (in str format) of called Quingo operation.

    Args:
        qu_src (str) :  the Quingo source containing the called operation.
        qg_func_name (str) : the name of the Quingo function.
    """
    qu_src = remove_comment(qu_src)

    search_string = r"\boperation\b\s+" + qg_func_name + r"\s*\((.*)\)\s*:(.*)\s*\{"

    op_def_components = re.search(search_string, qu_src)
    if op_def_components is None:
        raise ValueError(
            "Cannot find the operation ({}) in the given Quingo source "
            "file: \n```\n{}\n```".format(qg_func_name, qu_src)
        )

    ret_type = re.sub(r"\d", "", op_def_components.groups()[1].strip())
    if ret_type == "":
        ret_type = "unit"
    return ret_type


def gen_main_file(
    called_qg_fn: Path, called_qg_func: str, cl_entry_fn: Path, qg_params: tuple
):
    """This function generates the main function required to perform
    compilation. A new file named 'main_<qg_filename>' under the
    `<build_dirname>` directory is generated to allocate this main
    function.

    Args:
        called_qg_fn (Path) :
            the path of the Quingo file which contains the quantum
            function called by the host program.
        called_qg_func (str) :
            the name of the Quingo function.
        cl_entry_fn (Path) :
            compilation entry file, which contains the generated  main function
        args (list):
            a variable length of parameters passed to the quantum function
    """
    logger.setLevel(logging.WARNING)
    import_stmt = "import {}".format(called_qg_fn.stem)

    with called_qg_fn.open("r", encoding="utf-8") as qu_src_file:
        qu_src = qu_src_file.read()

    main_func_str = gen_main_func(qu_src, called_qg_func, qg_params)
    try:
        content = import_stmt + "\n" + main_func_str
        cl_entry_fn.write_text(content)
    except:
        raise IOError("Cannot write the file: ", cl_entry_fn)


def gen_main_func(qu_src: str, called_op: str, qg_params: tuple):
    """This function is used to generate string version of the main
    function used to call quingo.

    Args:
        called_op (str) : the name of called Quingo operation._name`.
    """

    var_name_list = []
    arg_str_list = []

    logger.debug(
        "calling function '{}' with parameters: {}".format(called_op, qg_params)
    )

    if len(qg_params) != 0:
        for i, arg in enumerate(qg_params):
            var_name, var_def_str = dt.conv_arg_to_qg_str(i, arg)
            # print("var_name: ", var_name)
            # print("var_def_str: ", var_def_str)
            var_name_list.append(var_name)
            arg_str_list.append(var_def_str)

    # qu_src = None
    # with called_qg_fn.open("r", encoding="utf-8") as qu_src_file:
    #     qu_src = qu_src_file.read()
    ret_type = get_ret_type(qu_src, called_op)

    str_params = ", ".join(var_name_list)
    arg_strs = "\n".join(arg_str_list)
    if len(arg_strs) > 0:
        arg_strs = "\n" + arg_strs

    str_ret = "" if ret_type == "unit" else "return "

    func_call = "{ret}{func_name}({parameters});".format(
        ret=str_ret, func_name=called_op, parameters=str_params
    )

    func_str = "\noperation main() : {} {{{}\n{}\n}}".format(
        ret_type, arg_strs, func_call
    )

    return func_str



import sympy as sp
from pathlib import Path
import numpy as np


def ensure_path(fn) -> Path:
    assert isinstance(fn, (str, Path))
    if isinstance(fn, str):
        fn = Path(fn).resolve()
    return fn


def is_number(s):
    try:
        float(s)
        return True
    except ValueError:
        return False


def singleton(cls):
    _instance = {}

    def inner():
        if cls not in _instance:
            _instance[cls] = cls()
        return _instance[cls]

    return inner


def number_distance(a, b):
    """calculates the distance between two complex numbers or two sympy numbers."""
    if isinstance(a, complex) and isinstance(b, complex):
        return (a.real - b.real) ** 2 + (a.imag - b.imag) ** 2

    if hasattr(a, "evalf") and hasattr(b, "evalf"):
        return (a.evalf() - b.evalf()) ** 2

    raise ValueError("unsupported types for dist: {} and {}".format(type(a), type(b)))


import numpy as np


def state_fidelity(state_a: np.ndarray, state_b: np.ndarray):
    """
    Calculate the state fidelity between two quantum states.

    Parameters:
    state_a (np.ndarray): The first quantum state.
    state_b (np.ndarray): The second quantum state.

    Returns:
    float: The state fidelity between the two quantum states.
    """
    return np.vdot(state_a, state_b)



from .backend_hub import BackendType, Backend_hub
from quingo.core.exe_config import ExeConfig


class If_backend:
    """The interface class for Quingo backends.

    Any backend should inherit from this interface.
    """

    def __init__(self, be_type: BackendType):
        self._be_type = be_type

    def get_type(self):
        return self._be_type

    def get_qisa(self):
        return Backend_hub().get_qisa(self._be_type)

    def is_simulator(self):
        return Backend_hub().is_simulator(self._be_type)

    def upload_program(self, program):
        raise NotImplementedError

    def execute(self, exe_config: ExeConfig):
        raise NotImplementedError



from __future__ import annotations
from numpy.typing import NDArray
from typing import List, Union


from quingo.utils import ensure_path
from .backend_hub import BackendType
from .if_backend import If_backend
from quingo.core.exe_config import ExeMode, ExeConfig
from pyqcisim.simulator import PyQCISim


class PyQCISim_quantumsim(If_backend):
    """A functional QCIS simulation backend using PyQCISim and QuantumSim."""

    def __init__(self):
        super().__init__(BackendType.QUANTUM_SIM)
        self.sim = PyQCISim()

    def upload_program(self, prog_fn):
        """
        Upload assembly or binary program to the simulator.

        Args:
            prog_fn: the name of the assembly or binary file.
        """
        prog_fn = ensure_path(prog_fn)

        program = prog_fn.open("r").read()
        self.sim.compile(program)

    def execute(self, exe_config: ExeConfig) -> Union[List | NDArray]:
        """Execute the given quantum circuit.
        Args:
          - mode (str): the simulation mode to use:

        The number of shots is specified in exe_config.num_shots, which is only valid for
          ExeMode.SimShots.
        """
        if exe_config.mode == ExeMode.SimShots:
            raw_res = self.sim.simulate("one_shot", exe_config.num_shots)
            return raw_res

        if exe_config.mode == ExeMode.SimFinalResult:
            raw_res = self.sim.simulate("final_result")
            return raw_res

        if exe_config.mode == ExeMode.SimStateVector:
            raw_res = self.sim.simulate("state_vector")
            return raw_res

        raise ValueError(
            "Unsupported execution mode ({}) for quantumsim.".format(exe_config.mode)
        )



import enum


class Qisa(enum.Enum):
    QCIS = enum.auto()
    QUIET = enum.auto()
    eQASM = enum.auto()
    Quantify = enum.auto()


def get_qisa_name(qisa: Qisa):
    """Get the name of the qisa type which is used by the compiler"""
    assert isinstance(qisa, Qisa)

    if qisa == Qisa.QCIS:
        return "qcis"
    if qisa == Qisa.QUIET:
        return "quiets"
    if qisa == Qisa.eQASM:
        return "eqasm"
    if qisa == Qisa.Quantify:
        return "quantify"

    raise ValueError("unknown qisa type: {}".format(qisa))


def get_suffix(qisa: Qisa):
    """Get the suffix of the corresponding qisa file"""
    assert isinstance(qisa, Qisa)

    if qisa == Qisa.QCIS:
        return ".qcis"
    if qisa == Qisa.QUIET:
        return ".qi"
    if qisa == Qisa.eQASM:
        return ".eqasm"
    if qisa == Qisa.Quantify:
        return ".json"

    raise ValueError("unknown qisa type: {}".format(qisa))



from pathlib import Path
import shutil
import quingo.global_config as gc
import tempfile
from quingo.backend.backend_hub import BackendType
from quingo.backend.qisa import Qisa
from quingo.utils import ensure_path

DEBUG_MODE = False


def create_empty_dir(dir_path: Path):
    if dir_path.exists():
        shutil.rmtree(str(dir_path))
    dir_path.mkdir()


class Quingo_task:
    def __init__(self, called_qu_fn: Path, called_func: str, **kwargs) -> None:
        """
        Define a quingo task by specifying the quingo file and the entry function.

        Note, this task only stores the following five elements:
        - the path of the called Quingo file (called_qu_fn)
        - the name of the entry function (called_func)
        - the path of the build directory (build_dir)
        - debug_mode
        - the target qisa (qisa) if specified upon initialization
        - the target backend (backend) if specified upon initialization

        If `qisa` and `backend` are not specified, then the value of them will be infered
        when required.

        The `build_dir` is calculated at the first time it is called. If the debug mode is
        on, then the build_dir is under the same directory as the called Quingo file.
        Otherwise, the build_dir is a temporary directory.

        Except them, all other properties are calculated when they are called. By doing so,
        only a minimal amount of information is stored in the task object, which reduces the
        risk of inconsistency when using this object.
        """
        # file name and function name
        called_qu_fn = ensure_path(called_qu_fn)
        self._called_qu_fn = called_qu_fn
        self._called_func = called_func
        self._build_dir = None

        self._qubits_info = kwargs.get("qubits_info", None)
        self.debug_mode = kwargs.get("debug_mode", False)
        # qisa and backend
        self._qisa = kwargs.get("qisa", None)
        self._backend = kwargs.get("backend", BackendType.QUANTUM_SIM)

    @property
    def qubits_info(self):
        """The path of the qubits information file."""
        return self._qubits_info

    @property
    def cl_entry_fn(self):
        """Generate the name of the Quingo file that is used to contain the main function."""
        main_qu_stem = "main_" + self.called_qu_fn.stem + "_" + self.called_func
        _cl_entry_fn = (self.build_dir / main_qu_stem).with_suffix(gc.quingo_suffix)
        return _cl_entry_fn

    @property
    def called_func(self):
        """The name of the called function."""
        return self._called_func

    @property
    def called_qu_fn(self):
        """The path of the called Quingo file."""
        return self._called_qu_fn

    @property
    def build_dir(self):
        """The path of the build directory.

        - In Debug mode, the build directory is under the current directory.
        - Otherwise, the build directory is a temporary directory.
        """
        # since build_dir can be a temporary directory, we need to record it instead of
        # calculating it every time.
        if self._build_dir is not None:
            return self._build_dir

        if self.debug_mode:  # debug mode create `build` dir under the current dir
            self._build_dir = Path.cwd() / gc.build_dirname
            create_empty_dir(self._build_dir)

        else:
            self._build_dir = Path(tempfile.mkdtemp(prefix="quingo-"))

        return self._build_dir

    @property
    def qisa_type(self):
        """
        Infer the qisa type from the selected backend.
        """
        if self._qisa is not None:
            return self._qisa

        if self._backend in [
            BackendType.QUANTUM_SIM,
            BackendType.TEQUILA,
            BackendType.SYMQC,
            BackendType.XIAOHONG,
            BackendType.QUALESIM_QUANTUMSIM,
            BackendType.QUALESIM_TEQUILA,
        ]:
            return Qisa.QCIS

        if self._backend in [BackendType.QUANTIFY]:
            return Qisa.Quantify

        raise ValueError(f"Found unknown backend {self._backend}")

    @property
    def called_qu_dir(self):
        """The directory of the called Quingo file."""

        # use resolve() to handle relative path like `..`
        abs_called_qg_fn = Path(self.called_qu_fn).resolve()
        return abs_called_qg_fn.parent

    @property
    def include_dir(self):
        """
        Generate the include directories for the Quingo compiler.
        """
        _inc_dir = []
        _inc_dir.append(self.called_qu_dir)
        _inc_dir.append(self.build_dir)
        return _inc_dir



import logging
from logging.handlers import TimedRotatingFileHandler
import sys
import colorama as cm
import termcolor as tc

cm.init()

# Fore: BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE, RESET.
# Back: BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE, RESET.
# Style: DIM, NORMAL, BRIGHT, RESET_ALL


def quingo_info(arg, **kwargs):
    print(arg, **kwargs)


def quingo_msg(arg, **kwargs):
    print(tc.colored(arg, "green"), **kwargs)


def quingo_warning(arg, **kwargs):
    print(tc.colored(arg, "yellow"), **kwargs)


def quingo_err(arg, **kwargs):
    print(tc.colored(arg, "red"), **kwargs)


FORMATTER = logging.Formatter(
    "%(asctime)s %(name)s %(lineno)d(%(levelname)s): %(message)s", datefmt="%H:%M:%S"
)
# LOG_FILE = "my_app.log"


def get_console_handler():
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(FORMATTER)
    return console_handler


# def get_file_handler():
#     file_handler = TimedRotatingFileHandler(LOG_FILE, when='midnight')
#     file_handler.setFormatter(FORMATTER)
#     return file_handler


def get_logger(logger_name):
    logger = logging.getLogger(logger_name)
    # better to have too much log than not enough
    logger.setLevel(logging.DEBUG)
    logger.addHandler(get_console_handler())
    # logger.addHandler(get_file_handler())
    # with this pattern, it's rarely necessary to propagate the error up to parent
    logger.propagate = False
    return logger



import importlib
import enum
from quingo.backend.qisa import Qisa
from quingo.utils import singleton


class BackendType(enum.Enum):
    QUANTUM_SIM = enum.auto()
    TEQUILA = enum.auto()
    SYMQC = enum.auto()
    QUANTIFY = enum.auto()
    XIAOHONG = enum.auto()
    QUALESIM_TEQUILA = enum.auto()
    QUALESIM_QUANTUMSIM = enum.auto()


@singleton
class Backend_hub:
    def __init__(self):
        self.backends = {
            BackendType.QUANTUM_SIM: (
                "PyQCISim_quantumsim",  # backend class name
                "pyqcisim_quantumsim",  # module path
                True,  # is_simulator
                Qisa.QCIS,  # qisa
            ),
            BackendType.TEQUILA: (
                "PyQCISim_tequila",
                "pyqcisim_tequila",
                True,
                Qisa.QCIS,
            ),
            BackendType.QUANTIFY: (
                "quantify",
                "to be added the hardware driver",
                False,
                Qisa.Quantify,
            ),
            BackendType.SYMQC: (
                "IfSymQC",
                "symqc",
                True,
                Qisa.QCIS,
            ),
            BackendType.XIAOHONG: (
                "XiaoHong",
                "xiaohong",
                False,
                Qisa.QCIS,
            ),
            BackendType.QUALESIM_TEQUILA: (
                "QuaLeSim",
                "qualesim",
                True,
                Qisa.QCIS,
            ),
            BackendType.QUALESIM_QUANTUMSIM: (
                "QuaLeSim",
                "qualesim",
                True,
                Qisa.QCIS,
            ),
        }

    def support(self, backend_type):
        return backend_type in self.backends

    def import_be_module(self, module_path):
        """import the backend module based on the module_path.

        The module is imported here because the backend module may not be installed.
        We only import the module when it is needed.
        """
        prefix = "quingo.backend."
        full_module_path = prefix + module_path
        module = importlib.import_module(full_module_path)
        return module

    def is_simulator(self, backend_type):
        be_class_name, module_path, is_sim, qisa = self.backends[backend_type]
        return is_sim

    def get_qisa(self, backend_type):
        be_class_name, module_path, is_sim, qisa = self.backends[backend_type]
        return qisa

    def get_instance(self, backend_type):
        be_class_name, module_path, is_sim, qisa = self.backends[backend_type]
        return getattr(self.import_be_module(module_path), be_class_name)()



from quingo.core.exe_config import ExeConfig, ExeMode
from quingo.core.compile import compile
from quingo.core.manager import execute, call
from quingo.core.quingo_task import Quingo_task
from quingo.backend.backend_hub import BackendType
from quingo.backend.qisa import Qisa



from pathlib import Path

qgrtsys_root_dir = Path(__file__).absolute().parent
mlir_compiler_config_path = qgrtsys_root_dir / "core" / "mlir_compiler_path.txt"

default_mlir_compiler_path = Path.home() / ".quingo" / "quingoc"

std_op_fn = "standard_operations.qu"
std_qfg_fn = "config-quingo.qfg"
std_op_full_path = qgrtsys_root_dir / "lib" / std_op_fn
std_qfg_full_path = qgrtsys_root_dir / "lib" / std_qfg_fn


quingo_suffix = ".qu"
eqasm_suffix = ".eqasm"
qcis_suffix = ".qcis"
quantify_suffix = ".json"
res_bin_suffix = ".bin"
build_dirname = "build"

shared_mem_start_addr = 0x000
shared_mem_size = 0x100000

QU_BOOL_SIZE = 1
QU_INT_SIZE = 4
QU_PTR_SIZE = 4
QU_DOUBLE_SIZE = 4

allowed_primitive_types = ["int", "bool", "double", "time"]
allowed_python_types = [
    "int",
    "bool",
    "float",
    "float32",
    "float64",
    "list",
    "tuple",
    "ndarray",
    "Time",
]

endian = "little"



from quingo.backend.backend_hub import BackendType
from quingo.backend.pyqcisim_tequila import PyQCISim_tequila
from quingo.backend.pyqcisim_quantumsim import PyQCISim_quantumsim
from quingo.backend.qualesim import QuaLeSim
from quingo.backend.symqc import IfSymQC
from quingo.backend.qisa import Qisa



from quingo.utils import ensure_path
from .backend_hub import BackendType
from .if_backend import If_backend
from quingo.core.exe_config import ExeConfig, ExeMode
from pyqcisim.simulator import PyQCISim
import numpy as np


class PyQCISim_tequila(If_backend):
    """A functional QCIS simulation backend using PyQCISim and Tequila."""

    def __init__(self):
        super().__init__(BackendType.TEQUILA)
        self.sim = PyQCISim()
        self.sim.setBackend("tequila")
        self.res = None

    def upload_program(self, prog_fn):
        prog_fn = ensure_path(prog_fn)

        program = prog_fn.open("r").read()
        self.sim.compile(program)

    def execute(self, exe_config: ExeConfig):
        """Execute the given quantum circuit.
        Args:
          - exe_config (ExeConfig): the configuration used to perform simulation by SymQC.

        The number of shots is specified in exe_config.num_shots, which is only valid for
          ExeMode.SimShots.
        """

        if exe_config.mode == ExeMode.SimShots:
            return self.sim.simulate("one_shot", exe_config.num_shots)

        if exe_config.mode == ExeMode.SimFinalResult:
            return self.sim.simulate("final_result")

        if exe_config.mode == ExeMode.SimStateVector:
            names, nd_array_values = self.sim.simulate("state_vector")
            return (names, nd_array_values)

        raise ValueError(
            "Unsupported execution mode ({}) for TEQUILA.".format(exe_config.mode)
        )



class Time:
    def __init__(self, value: int = None, unit: str = None) -> None:
        self.value = value
        self.unit = unit

    def __str__(self) -> str:
        if self.value is None or self.unit is None:
            raise ValueError("Time value or unit is not set.")
        return f"{self.value}{self.unit}"

    def set_value_and_unit(self, value: int, unit: str) -> None:
        self.value = value
        self.unit = unit



import re
import struct
import numpy

import quingo.global_config as gc
from quingo.core.data import Time
import logging
from quingo.core.quingo_logger import get_logger
from typing import Tuple

logger = get_logger((__name__).split(".")[-1])
logger.setLevel(logging.WARNING)


def check_if_param_type(arg) -> str:
    arg_type = type(arg).__name__

    if arg_type not in gc.allowed_python_types:
        raise ValueError(
            "Unsupported type ({}) of argument ({}) has been"
            " passed to Quingo.".format(arg_type, arg)
        )

    if isinstance(arg, float):
        arg_type = "double"

    if isinstance(arg, numpy.ndarray):
        arg_type = "list"

    if isinstance(arg, Time):
        arg_type = "time"

    return arg_type


def get_list_type_str(arg) -> str:
    assert isinstance(arg, list)

    # TODO: currently, we cannot infer the type of a list with zero elements
    #       This should be fixed in the future.
    if len(arg) == 0:
        raise NotImplementedError(
            "Currently cannot pass a list with a length of zero to Quingo."
        )

    # Quingo only accepts a list of elements with the same type
    if not all(isinstance(x, type(arg[0])) for x in arg):
        raise ValueError(
            "Interface parameter error: a list ({}) of elements of"
            " different types is passed to Quingo.".format(arg)
        )

    list_elem_type = check_if_param_type(arg[0])

    if list_elem_type == "list":
        type_str = get_list_type_str(arg[0]) + "[]"
    elif list_elem_type == "tuple":
        type_str = get_tuple_type_str(arg[0]) + "[]"
    else:
        type_str = list_elem_type + "[]"

    return type_str


def get_tuple_type_str(arg) -> str:
    assert isinstance(arg, tuple)

    type_str = "("

    for elem in arg:
        tuple_elem_type = check_if_param_type(elem)

        if tuple_elem_type == "list":
            type_str += get_list_type_str(elem) + ", "
        elif tuple_elem_type == "tuple":
            type_str += get_tuple_type_str(elem) + ", "
        else:
            type_str += tuple_elem_type + ", "

    type_str = type_str[:-2] + ")"

    return type_str


def conv_arg_to_qg_str(i: int, arg) -> Tuple[str, str]:
    arg_def_str = ""

    var_prefix = "var"

    arg_type = check_if_param_type(arg)

    if arg_type in gc.allowed_primitive_types:
        # Give a name to this Quingo variable
        var_name = "{}{}_{}".format(var_prefix, i, arg_type)
        # Quingo variable declaration
        arg_def_str = "{} {};\n".format(arg_type, var_name)
        # assign the value
        arg_def_str += "{} = {};\n".format(var_name, str(arg).lower())

    elif arg_type == "list":
        if isinstance(arg, numpy.ndarray):
            arg = arg.tolist()

        concrete_list_type = get_list_type_str(arg)

        # Give a name to this Quingo variable
        var_name = "{}{}_arr".format(var_prefix, i)
        # Quingo variable declaration
        arg_def_str = "{} {};\n".format(concrete_list_type, var_name)

        # assign the value
        array_content = str(arg).lower().replace("[", "{").replace("]", "}")
        arg_def_str += "{} = {};\n".format(var_name, array_content)

    elif arg_type == "tuple":
        concrete_tuple_type = get_tuple_type_str(arg)

        # Give a name to this Quingo variable
        var_name = "{}{}_tpl".format(var_prefix, i)
        # Quingo variable declaration
        arg_def_str = "{} {};\n".format(concrete_tuple_type, var_name)
        # assign the value
        tuple_content = str(arg).lower().replace("[", "{").replace("]", "}")
        arg_def_str += "{} = {};\n".format(var_name, tuple_content)

    return var_name, arg_def_str


def get_tuple_ele_type(tuple_str):
    tuple_str = tuple_str.replace(" ", "")

    if len(tuple_str) < 2 or tuple_str[0] != "(" or tuple_str[-1] != ")":
        raise ValueError("Provided type string ({}) is not a tuple!".format(tuple_str))

    return split_comma_sep_types(tuple_str[1:-1])


def get_end_square_addr(type_str, start_addr):
    end_addr = start_addr

    while end_addr < len(type_str) - 1:
        if type_str[end_addr : end_addr + 1] != "[]":
            break

    return end_addr


def split_comma_sep_types(comma_sep_types):
    comma_sep_types = comma_sep_types.replace(" ", "")

    type_list = []

    end_square_addr = 0

    if comma_sep_types[0] == "(":  # tuple
        right_brace_addr = get_paired_brace(comma_sep_types, 0)
        end_square_addr = get_end_square_addr(comma_sep_types, right_brace_addr)
    else:
        if comma_sep_types.startswith("int"):
            end_square_addr = get_end_square_addr(comma_sep_types, len("int") - 1)
        elif comma_sep_types.startswith("bool"):
            end_square_addr = get_end_square_addr(comma_sep_types, len("bool") - 1)

    comma_addr = comma_sep_types.find(",", end_square_addr)

    if comma_addr == -1:
        type_list.append(comma_sep_types)
    else:
        type_list.append(comma_sep_types[0:comma_addr])

        new_string = comma_sep_types[comma_addr + 1 :]

        type_list.extend(split_comma_sep_types(new_string))

    return type_list


def get_paired_brace(type_str, start_addr):
    assert type_str[start_addr] == "("

    level = 0

    i = start_addr
    for i in range(start_addr, len(type_str)):
        if type_str[i] == "(":
            level += 1

        if type_str[i] == ")":
            level -= 1

        if level == 0:
            return i

    raise ValueError("Found unpaired braces in the type string ({}).".format(type_str))


def get_outer_type(type_str):
    type_str = type_str.replace(" ", "")

    length = len(type_str)

    if length > 3 and type_str[length - 2 : length] == "[]":
        return "list"

    elif len(type_str) > 2 and type_str[0] == "(" and type_str[-1] == ")":
        return "tuple"

    elif type_str == "int" or type_str == "bool" or type_str == "double":
        return type_str
    else:
        raise ValueError("Unrecognized type string ({}) is provided.".format(type_str))


class Data_transfer:
    """Data_transfer serves as the backbone module."""

    def __init__(self, **kwargs):
        self.data_block = None
        self.head = 0

    def set_data_block(self, bin_block):
        """This function set the data block to be decoded, and reset the
        decoder pointer.

        Args:
            bin_block (binary) : the binary data block returned by the quantum kernel.
        """
        self.data_block = bin_block
        self.head = 0

    def _byte_2_int(self, byte_data):
        return int.from_bytes(byte_data, byteorder=gc.endian, signed=True)

    def _byte_2_bool(self, byte_data):
        return bool.from_bytes(byte_data, byteorder=gc.endian, signed=True)

    def _byte_2_double(self, byte_data):
        return struct.unpack("<f", byte_data)[0]

    def _check_ptr_range(self, head, tail):
        if head < 0:
            raise ValueError("Error: found negative pointer ({})!", format(head))

        if tail > len(self.data_block):
            raise ValueError(
                "The pointer ({}) to the data block exceeds the "
                "maximum size ({}).".format(tail, len(self.data_block))
            )

    def _get_int(self, head):
        """This function read an integer from the data block, and move the
        decode pointer forward with a step size equal to the integer size.
        """

        new_head = head + gc.QU_INT_SIZE
        self._check_ptr_range(head, new_head)
        int_data = self.data_block[head:new_head]
        return self._byte_2_int(int_data), new_head

    def _get_bool(self, head):
        """This function read a bool from the data block, and move the decode
        pointer forward with a step size equal to the bool size.
        """

        new_head = head + gc.QU_BOOL_SIZE
        self._check_ptr_range(head, new_head)
        int_data = self.data_block[head:new_head]
        return self._byte_2_bool(int_data), new_head

    def _get_double(self, head):
        """This function read a double from the data block, and move the decode
        pointer forward with a step size equal to the double size.
        """

        new_head = head + gc.QU_DOUBLE_SIZE
        self._check_ptr_range(head, new_head)
        int_data = self.data_block[head:new_head]
        return self._byte_2_double(int_data), new_head

    def _get_pointer(self, head):
        new_head = head + gc.QU_PTR_SIZE
        self._check_ptr_range(head, new_head)
        int_data = self.data_block[head:new_head]
        return self._byte_2_int(int_data), new_head

    def bin_to_pydata(self, type_str, start_addr=0x600):
        logger.debug(
            "The binary data is going to be decoded according to the type: ", type_str
        )
        res, head = self.conv_qg_bin_to_py_data(type_str, start_addr)
        return res

    def conv_qg_bin_to_py_data(self, type_str, head):
        outer_type = get_outer_type(type_str)

        if outer_type == "int":
            value, head = self._get_int(head)

        elif outer_type == "bool":
            value, head = self._get_bool(head)

        elif outer_type == "double":
            value, head = self._get_double(head)

        elif outer_type == "tuple":
            value, head = self.conv_qg_bin_to_py_tuple(type_str, head)

        elif outer_type == "list":
            value, head = self.conv_qg_bin_to_py_array(type_str, head)

        else:
            raise ValueError("Undefined types ({}) found.".format(arg))

        return value, head

    def conv_qg_bin_to_py_tuple(self, type_str, head):
        type_list = get_tuple_ele_type(type_str)

        value = ()
        for ele_type in type_list:
            ele_value, head = self.conv_qg_bin_to_py_data(ele_type, head)
            value += (ele_value,)
        return value, head

    def conv_qg_bin_to_py_array(self, type_str, head):
        arr_head_offset, newhead = self._get_pointer(head)
        arr_head = head + arr_head_offset
        ele_type = type_str[:-2]
        value = []
        arr_length, arr_head = self._get_int(arr_head)
        for i in range(arr_length):
            sub_arr, arr_head = self.conv_qg_bin_to_py_data(ele_type, arr_head)
            value.append(sub_arr)
        head = newhead
        return value, head



import pathlib
import re
import math
import requests
import json
import distutils.spawn
import subprocess
import platform
import zipfile
import sys
import shutil
import tempfile
import datetime
import tqdm
from pathlib import Path

quingoc_owner = "quingo"
quingoc_release_repo = "quingoc-release"
lastest_quingoc_release_url = (
    "https://gitee.com/api/v5/repos/{owner}/{repo}/releases/latest".format(
        owner=quingoc_owner, repo=quingoc_release_repo
    )
)

os_name_list = ["Linux", "Windows", "Darwin"]
os_dl_suffix = {"Linux": ".sh", "Windows": ".zip", "Darwin": ".dmg"}


def backup_compiler(quingoc_path):
    backup_path = quingoc_path.parent / "quingoc_backup"

    if not backup_path.exists():
        backup_path.mkdir(exist_ok=True)

    backup_file = quingoc_path.name + datetime.datetime.now().strftime("%F")

    shutil.move(quingoc_path, backup_path / backup_file)

    print('Local quingoc has been backup at "{}".'.format(backup_path))


def set_path_env_on_Linux(install_path):
    """Set path environment to contain the milr quingo compiler on Linux."""

    ret_value = subprocess.run(
        "echo $PATH",
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
        shell=True,
    )

    if ret_value.returncode != 0:
        raise RuntimeError("Failed to retrieve system path using 'echo $PATH'.")

    if str(install_path.absolute()) in ret_value.stdout:
        return

    bash_profile = Path.home() / ".bash_profile"
    bashrc = Path.home() / ".bashrc"
    shell_cmd = "if [ -f \"{bashrc_path}\" ]\nthen\n \
                    echo 'export PATH={quingoc_dir_path}:$PATH' >> {bashrc_path}\n \
                elif [ -f \"{bash_profile_path}\" ]\nthen\n \
                    echo 'export PATH={quingoc_dir_path}:$PATH' >> {bash_profile_path}\n \
                fi\n"
    set_env_cmd = shell_cmd.format(
        bashrc_path=bashrc,
        bash_profile_path=bash_profile,
        quingoc_dir_path=install_path,
    )

    ret_value = subprocess.run(
        set_env_cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
        shell=True,
    )

    if ret_value.returncode != 0:
        raise RuntimeError(
            'Failed add "{}" to path environment with the'
            "following error: {}".format(install_path, ret_value.stderr)
        )


def install_on_Linux(mlir_compiler_path, old_version_path=None):
    """Install quingo compiler on Linux."""

    mlir_compiler_path.chmod(0o744)

    if old_version_path is not None:
        mlir_compiler_install_dir = old_version_path.parent
    else:
        mlir_compiler_install_dir = Path.home() / ".local"

    mlir_compiler_install_cmd = '"{}" --prefix="{}" --exclude-subdir'.format(
        mlir_compiler_path, mlir_compiler_install_dir
    )

    ret_value = subprocess.run(
        mlir_compiler_install_cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
        shell=True,
    )
    if ret_value.returncode != 0:
        raise RuntimeError(
            "Failed to install lastest quingo compiler with the"
            "following error: {}".format(ret_value.stderr)
        )

    if old_version_path is not None:
        mlir_compiler_exec_path = mlir_compiler_install_dir
        shutil.copy(
            str(mlir_compiler_install_dir / "bin" / "quingoc"),
            str(mlir_compiler_exec_path),
        )
    else:
        mlir_compiler_exec_path = mlir_compiler_install_dir / "bin"

    set_path_env_on_Linux(mlir_compiler_exec_path)

    print(
        'Successfully installed the Lastest quingoc at "{}".'.format(
            mlir_compiler_exec_path
        )
    )


def set_path_on_Windows(install_path):
    """Set path environment to contain the milr quingo compiler on Windows."""
    # Nothing to do, quingo-runtime will search the default install path
    # Default path is Path.home() / '.quingo'

    return


def install_on_Windows(mlir_compiler_path, old_version_path=None):
    """Install quingo compiler on Windows."""

    if old_version_path is not None:
        mlir_compiler_install_path = old_version_path.parent
    else:
        mlir_compiler_install_path = Path.home() / ".quingo"

    zip_file = zipfile.ZipFile(mlir_compiler_path)
    zip_file.extractall(mlir_compiler_install_path)
    zip_file.close()

    mlir_compiler_extract_path = mlir_compiler_install_path / mlir_compiler_path.stem

    files = [
        file for file in mlir_compiler_extract_path.glob("bin/*") if file.is_file()
    ]
    for file in files:
        shutil.copy(file, mlir_compiler_install_path)

    shutil.rmtree(mlir_compiler_extract_path)

    set_path_on_Windows(mlir_compiler_install_path)

    print(
        'Successfully installed the Lastest quingoc at "{}".'.format(
            mlir_compiler_install_path
        )
    )


def set_path_env_on_Darwin(install_path):
    """Set path environment to contain the milr quingo compiler on Darwin."""

    set_path_env_on_Linux(install_path)


def install_on_Darwin(mlir_compiler_path, old_version_path=None):
    """Install quingo compiler on Darwin."""

    if old_version_path is not None:
        mlir_compiler_install_path = old_version_path.parent
        mlir_compiler_exec_path = mlir_compiler_install_path
    else:
        mlir_compiler_install_path = Path.home() / ".local"
        mlir_compiler_exec_path = mlir_compiler_install_path / "bin"

    if not mlir_compiler_exec_path.exists():
        mlir_compiler_exec_path.mkdir(exist_ok=True)

    try:
        # mount dmg file
        mount_cmd = "hdiutil attach " + str(mlir_compiler_path)

        ret_value = subprocess.run(
            mount_cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
            shell=True,
        )
        if ret_value.returncode != 0:
            raise RuntimeError(
                "Failed to mount lastest quingo compiler dmg file with the"
                "following error: {}".format(ret_value.stderr)
            )

        mlir_compiler_exec_file = (
            pathlib.Path("/Volumes/" + mlir_compiler_path.stem)
            / "quingoc.app"
            / "Contents"
            / "Resources"
            / "bin"
            / "quingoc"
        )

        shutil.copy(str(mlir_compiler_exec_file), str(mlir_compiler_exec_path))

    except Exception as e:
        raise RuntimeError(
            "Failed to copy quingo compiler {} to '{}' with the "
            "following error: {}".format(
                mlir_compiler_exec_file, mlir_compiler_exec_path, e
            )
        )

    finally:
        # umount dmg file
        umount_cmd = "hdiutil detach " + str("/Volumes/" + mlir_compiler_path.stem)

        ret_value = subprocess.run(
            umount_cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
            shell=True,
        )
        if ret_value.returncode != 0:
            raise RuntimeError(
                "Failed to umount lastest quingo compiler dmg file with the"
                "following error: {}".format(ret_value.stderr)
            )

    set_path_env_on_Darwin(mlir_compiler_exec_path)

    print(
        'Successfully installed the Lastest quingoc at "{}".'.format(
            mlir_compiler_exec_path
        )
    )


def install_compiler(os_name, mlir_compiler_path, old_version_path=None):
    if old_version_path is not None:
        backup_compiler(old_version_path)

    if os_name == "Windows":
        install_on_Windows(mlir_compiler_path, old_version_path)
    if os_name == "Linux":
        install_on_Linux(mlir_compiler_path, old_version_path)
    if os_name == "Darwin":
        install_on_Darwin(mlir_compiler_path, old_version_path)


def download_compiler(os_name, tmp_dir_name):
    assert os_name in os_name_list

    dl_file_suffix = os_dl_suffix[os_name]

    try:
        latest_release = requests.get(lastest_quingoc_release_url).text
        release_info = json.loads(latest_release)

    except Exception as e:
        raise RuntimeError(
            "Failed to parse information retrieved from gitee with the "
            "following error: {}".format(e)
        )

    assets = release_info["assets"]
    quingoc_asset = None
    for asset in assets:
        if "name" in asset and asset["name"].endswith(dl_file_suffix):
            quingoc_asset = asset
            break
    if quingoc_asset is None:
        raise RuntimeError(
            "Failed to get quingo compiler release package for {} platform.".format(
                os_name
            )
        )

    quingoc_url = quingoc_asset["browser_download_url"] + "/"

    try:
        quingoc_response = requests.get(quingoc_url, stream=True)
    except Exception as e:
        raise RuntimeError(
            'Failed to download package "{}" from gitee with the '
            "following error: {}".format(quingoc_asset["name"], e)
        )

    mlir_compiler_path = tmp_dir_name / quingoc_asset["name"]
    with mlir_compiler_path.open("wb") as tmp_dl_file:
        print("Downloading quingoc from {} ".format(quingoc_url))
        for data in tqdm.tqdm(
            iterable=quingoc_response.iter_content(1024),
            desc="progress",
            unit="KB",
        ):
            tmp_dl_file.write(data)

    return mlir_compiler_path


def download_and_install_latest_quingoc(old_version_path=None):
    """Download the latest mlir quingo compiler."""
    os_name = platform.system()
    if os_name not in ["Linux", "Windows", "Darwin"]:
        raise SystemError(
            "Currently pip installation does not support {}".format(os_name)
        )

    tmp_dir = tempfile.TemporaryDirectory()
    tmp_dir_path = Path(tmp_dir.name)

    mlir_compiler_path = download_compiler(os_name, tmp_dir_path)
    install_compiler(os_name, mlir_compiler_path, old_version_path)


def get_lastest_version():
    """Get lastest quingo compiler version"""

    os_name = platform.system()
    assert os_name in os_name_list

    dl_file_suffix = os_dl_suffix[os_name]

    try:
        latest_release = requests.get(lastest_quingoc_release_url).text
        release_info = json.loads(latest_release)

    except Exception as e:
        raise RuntimeError(
            "Failed to parse information retrieved from gitee with the "
            "following error: {}".format(e)
        )

    assets = release_info["assets"]
    quingoc_asset = None
    for asset in assets:
        if "name" in asset and asset["name"].endswith(dl_file_suffix):
            quingoc_asset = asset
            break
    if quingoc_asset is None:
        raise RuntimeError(
            "Failed to get quingo compiler release package for {} platform.".format(
                os_name
            )
        )

    find_version = re.search(r"\d+\.\d+\.\d+", quingoc_asset["name"])
    if find_version is not None:
        lastest_version = find_version.group()
    else:
        raise RuntimeError(
            'Failed to get lastest version of quingo compiler from package "{}".'.format(
                quingoc_asset["name"]
            )
        )

    return lastest_version


def get_current_version(quingoc_path):
    """Get current quingo compiler version"""

    get_current_version_cmd = "{} --version".format(quingoc_path)

    ret_value = subprocess.run(
        get_current_version_cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
        shell=True,
    )
    if ret_value.returncode != 0:
        raise RuntimeError(
            "Failed to get local quingo compiler version with the"
            "following error: {}".format(ret_value.stderr)
        )

    find_version = re.search(r"\d+\.\d+\.\d+", ret_value.stdout.split("\n")[0])
    if find_version is not None:
        current_version = find_version.group()
    else:
        raise RuntimeError(
            "Failed to get version of local quingo compiler from {}.".format(
                quingoc_path
            )
        )

    return current_version


def check_update(quingoc_path):
    """Check local quingo compiler whether is latest version"""

    current_version = get_current_version(quingoc_path)
    lastest_version = get_lastest_version()

    if current_version == lastest_version:
        print(
            "Local quingoc is already the lastest version {}.".format(lastest_version)
        )
        return False
    else:
        print(
            "Local quingoc version ({}) is behind lastest version ({}). Update now.".format(
                current_version, lastest_version
            )
        )
        return True


if __name__ == "__main__":
    default_path = Path.home() / ".quingo"
    quingoc_path = distutils.spawn.find_executable("quingoc", str(default_path))

    if quingoc_path is None:
        quingoc_path = distutils.spawn.find_executable("quingoc")

    if quingoc_path is None:
        download_and_install_latest_quingoc()
    else:
        if check_update(quingoc_path):
            download_and_install_latest_quingoc(pathlib.Path(quingoc_path))



from __future__ import annotations
from numpy.typing import NDArray
from typing import List, Union
import numpy as np
import array
import sympy as sp

from pathlib import Path

from quingo.core.exe_config import ExeConfig, ExeMode
from quingo.core.quingo_task import Quingo_task
from quingo.core.compile import compile
from quingo.backend.backend_hub import BackendType, Backend_hub


def verify_backend_config(backend: BackendType, exe_config: ExeConfig) -> bool:
    """Check if the combination of backend and execution configuration is valid."""
    if (backend == BackendType.XIAOHONG) and (exe_config.mode != ExeMode.RealMachine):
        return False

    if backend == BackendType.QUANTIFY:
        return False
    return True


def execute(
    qasm_fn: Path, be_type: BackendType, exe_config: ExeConfig = ExeConfig()
) -> Union[List | NDArray]:
    """Execute the quingo task on the specified backend and return the result."""

    if not verify_backend_config(be_type, exe_config):
        raise ValueError(
            "Error configuration {} on the backend {}".format(str(exe_config), backend)
        )

    backend = Backend_hub().get_instance(be_type)
    backend.upload_program(qasm_fn)
    result = backend.execute(exe_config)
    if exe_config.mode == ExeMode.SimStateVector:
        names, array_values = result
        if len(names) == 0:
            return ([], 1)

        if isinstance(array_values, list):
            if len(array_values) == 0:
                return ([], 1)

            array_values = np.array(array_values)

        elif isinstance(array_values, array.array):
            array_values = np.array(array_values)

        elif isinstance(array_values, sp.Matrix):
            array_values = np.array(array_values).astype(np.complex64)

        else:
            assert isinstance(array_values, np.ndarray)

        array_values = array_values.flatten()
        return (names, array_values)

    return result


def call(
    task: Quingo_task,
    params: tuple,
    be_type: BackendType = BackendType.QUANTUM_SIM,
    exe_config: ExeConfig = ExeConfig(),
    config_fn="",
):
    """Execute the quingo task on the specified backend and return the result."""

    qasm_fn = compile(task, params, config_file=config_fn)
    return execute(qasm_fn, be_type, exe_config)



from pathlib import Path
from quingo.core.quingo_logger import quingo_err, quingo_info
import quingo.global_config as gc
import distutils.spawn
import requests
import json


def get_text(link):
    r = requests.get(link)
    return r.text


def set_compiler_path(path_str):
    path = Path(path_str)
    if path_str.startswith("~"):
        path = path.expanduser()
    if not path.exists():
        quingo_err("cannot find the compiler at {}".format(path_str))
        return False

    path_file = gc.mlir_compiler_config_path

    with path_file.open("w") as f:
        f.write(str(path))

    return True


def download_latest_quingoc():
    latest_release_url = (
        "https://gitee.com/api/v5/repos/{owner}/{repo}/releases/latest".format(
            owner="quingo", repo="quingoc-release"
        )
    )

    try:
        latest_release = get_text(latest_release_url)
        release_info = json.loads(latest_release)
        assets = release_info["assets"]
        quingoc_asset = None
        for asset in assets:
            if "name" in asset and asset["name"].startswith("quingoc"):
                quingoc_asset = asset
                break
        if quingoc_asset is None:
            quingo_err("failed to download quingoc from gitee.")
            return False

        quingoc_url = (
            quingoc_asset["browser_download_url"] + "/" + quingoc_asset["name"]
        )
        quingoc_response = requests.get(quingoc_url)

    except Exception as e:
        raise RuntimeError(
            "Failed to parse information retrieved from gitee with the "
            "following error: {}".format(e)
        )

    with gc.default_mlir_compiler_path.open("wb") as f:
        f.write(quingoc_response.content)

    gc.default_mlir_compiler_path.chmod(0o744)

    return True


def retrieve_compiler_path_from_file(config_file):
    compiler_path = None
    with config_file.open("r") as f:
        compiler_path = f.read()
        if not Path(compiler_path).exists():
            quingo_err("The compiler path is incorrect: {}".format(compiler_path))
            return None
    return compiler_path


def get_mlir_path():
    if gc.mlir_compiler_config_path.exists():
        read_path = retrieve_compiler_path_from_file(gc.mlir_compiler_config_path)
        if read_path is not None:
            return read_path

    # failed to read the path from the file, find the compiler executable from default path
    quingoc_path = distutils.spawn.find_executable(
        "quingoc", str(gc.default_mlir_compiler_path.parent)
    )
    if quingoc_path is not None:
        return quingoc_path

    # find the compiler executable from the system directly
    quingoc_path = distutils.spawn.find_executable("quingoc")

    if quingoc_path is None:
        quingo_info(
            "Cannot find the compiler.\n"
            "  To resolve this problem, you can install quingoc with two ways:\n"
            '  1. run the following command "python -m quingo.install_quingoc"\n'
            "  2. Dowload quingoc from https://gitee.com/quingo/quingoc-release/"
            "releases and save it at a directory in the system path \n"
            "or configure its path by calling this method inside python:\n"
            "     `quingo.set_mlir_compiler_path(<path-to-quingoc>)`"
        )
        raise RuntimeError(
            "Cannot find the mlir-based quingoc compiler in the system path."
        )

    return quingoc_path


def set_mlir_compiler_path(mlir_path_str):
    set_compiler_path(mlir_path_str, True)
